"""
Contains functionality to transform TinyPseuCo into z3 formulas
"""

from functools import singledispatch, update_wrapper
from itertools import chain
import typing
from typing import List, Tuple, NamedTuple, Dict, Union
from enum import Enum, unique
import logging

from bidict import bidict
import z3
from z3 import ForAll, And, Or, Implies, Exists, Not

from pseuco_parser import (BaseExpression, UnaryExpression, BinaryExpression, UnaryOperator,
                           BinaryOperator, IntegerConstant, BooleanConstant, VariableUsage,
                           Statement, EmptyStatement, SimpleStatement, MultiStatement, ITE,
                           VariableAssignment, VariableDeclaration, PrintlnStatement, PseuCoType)


@unique
class MemoryModel(Enum):
    """Enum to select which memory model to use"""
    SequentialConsistency = "seqConsist"
    WeakConsistency = "weak"

class InstructionId(NamedTuple):
    agent_id: int
    statement_id: int


class OrderAssertion(NamedTuple):
    """Describes asserttion InstructionOrder(predecessor, statement)
    Implicitly marks statement and predecessor as Reachable
    """
    statement: InstructionId
    predecessor: InstructionId

class WriteSeenAssertion(NamedTuple):
    """Describes asserttion WriteSeen(writer, reader)
    Implicitly marks writer and reader as Reachable
    """
    writer: InstructionId
    reader: InstructionId
    variable: str

class ReadValueAssertion(NamedTuple):
    """Describes assertion ReadValue(instruction, variable, value)
    Implicitly marks instruction as Reachable
    """
    instruction: InstructionId
    variable: str
    value: int


def methoddispatch(func):
    """ Taken from
    https://stackoverflow.com/questions/24601722/how-can-i-use-functools-singledispatch-with-instance-methods
    Allows single dispatch for methods
    """
    dispatcher = singledispatch(func)

    def wrapper(*args, **kw):  # pylint: disable=missing-docstring
        return dispatcher.dispatch(args[1].__class__)(*args, **kw)
    wrapper.register = dispatcher.register
    update_wrapper(wrapper, func)
    return wrapper


class Z3Compiler:

    def __init__(self, timeout=20000) -> None:
        self.context = z3.Context()
        self.local_var_table: typing.Mapping[str, z3.ArithRef] = {}
        self.solver: z3.Solver = z3.Solver(ctx=self.context)
        self.solver.set(timeout=timeout)
        self.statement_enablers: List[z3.z3.ExprRef] = []
        self.global_reads: List[Tuple["Instruction", "PseuCoVar", z3.IntNumRef]] = []
        self.global_writes: List[Tuple["Instruction", "PseuCoVar", z3.IntNumRef]] = []
        self.parent_instructions: List["Instruction"] = []
        self.current_agent_id = 0
        self.agent_id2agent = {}
        self.direct_predecessors = {}
        self.PseuCoVar = None
        self.Instruction = None
        self.InstructionOrder = None
        self.Reachable = None
        self.ReadValue = None
        self.WriteValue = None
        self.ReadWrite = None

        self._instruction2instruction_id: Dict["Instruction", InstructionId] = bidict()
        self._current_agent_used_id_counter = 0
        self._timeout = timeout
        self._solver_stack_gauge = 0
        self._manually_added_assertions = []


    def _push(self):
        self.solver.push()
        self._solver_stack_gauge += 1

    def _pop(self):
        assert self._solver_stack_gauge > 0
        self.solver.pop()
        self._solver_stack_gauge -= 1

    def _pop_all(self):
        for _ in range(self._solver_stack_gauge):
            self.solver.pop()
        self._solver_stack_gauge = 0


    def make_global_read(self, varname: str,
                         instr: "Instruction") -> z3.z3.ArithRef:
        """Takes the variable name and where it is written
        :returns: a variable reperesenting the result of the read
        """
        read_value = z3.Const(str(self.current_agent_id) + str(instr),
                              z3.IntSort(ctx=self.context))
        variable = self.PseuCoVar.__getattribute__(varname)
        self.global_reads.append((instr, variable, read_value))
        return read_value

    def make_global_write(self,
                          varname: str,
                          instr: "Instruction",
                          written_value: z3.z3.ArithRef):
        """Prepares the addition of an global variable read constraint"""
        variable = self.PseuCoVar.__getattribute__(varname)
        self.global_writes.append((instr, variable, written_value))

    def create_local_pseuco_variable(self, name, var_type: PseuCoType):
        if var_type is PseuCoType.Int:
            new_var = z3.Int(str((self.current_agent_id, name)), ctx=self.context)
        elif var_type is PseuCoType.Boolean:
            new_var = z3.Bool(str((self.current_agent_id, name)), ctx=self.context)
        else:
            raise NotImplementedError("No support for %s" % str(var_type))
        assert name not in self.local_var_table, "no shadowing!"
        self.local_var_table[name] = new_var
        return new_var

    @methoddispatch
    def compile_expression(self, expression: BaseExpression) -> z3.z3.ExprRef:
        raise NotImplementedError("This does not work")

    @compile_expression.register(UnaryExpression)
    def _unary(self, unary_exp: UnaryExpression):
        if unary_exp.operator is UnaryOperator.Plus:
            return self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Minus:
            return -self.compile_expression(unary_exp.operand)
        elif unary_exp.operator is UnaryOperator.Not:
            return not self.compile_expression(unary_exp.operand)
        else:
            raise NotImplementedError("%s is not supported" %
                                      unary_exp.operator)

    @compile_expression.register(IntegerConstant)
    def _intconst(self, c: IntegerConstant) -> z3.z3.IntNumRef:
        return z3.IntVal(c.value, ctx=self.context)

    @compile_expression.register(BooleanConstant)
    def _boolconst(self, c: BooleanConstant) -> z3.z3.BoolRef:
        return z3.BoolVal(c.value, ctx=self.context)

    @compile_expression.register(BinaryExpression)
    def _binexp(self, binary_exp: BinaryExpression):
        lhs = self.compile_expression(binary_exp.lhs)
        rhs = self.compile_expression(binary_exp.rhs)
        op = binary_exp.operator
        if op is BinaryOperator.Plus:
            return lhs + rhs
        elif op is BinaryOperator.Minus:
            return lhs - rhs
        elif op is BinaryOperator.Mult:
            return lhs * rhs
        elif op is BinaryOperator.Div:
            return lhs / rhs
        elif op is BinaryOperator.And:
            return lhs and rhs
        elif op is BinaryOperator.Or:
            return lhs or rhs
        elif op is BinaryOperator.Equal:
            return lhs == rhs
        elif op is BinaryOperator.NEqual:
            return lhs != rhs
        elif op is BinaryOperator.Lt:
            return lhs < rhs
        elif op is BinaryOperator.Lte:
            return lhs <= rhs
        elif op is BinaryOperator.Gt:
            return lhs > rhs
        elif op is BinaryOperator.Gte:
            return lhs >= rhs

    @compile_expression.register(VariableUsage)
    def _varusage(self, var: VariableUsage):
        # NOTE: local_var_table is modified such that globals point to their current value
        return self.local_var_table[var.varname]

    def _create_and_order_instruction(self,
                                      statement: Statement) -> "Instruction":
        instr = self.Instruction.__getattribute__("a%d_%d" % (self.current_agent_id,
                                                              self._current_agent_used_id_counter))
        self._current_agent_used_id_counter += 1
        self._instruction2instruction_id[instr] = InstructionId(self.current_agent_id, statement.id)
        for parent in self.parent_instructions:
            self.solver.assert_and_track(self.InstructionOrder(parent, instr),
                                         "%s < %s" % (str(parent), str(instr)))
        self.parent_instructions = [instr]
        self.solver.assert_and_track(And(*self.statement_enablers, self.context) ==
                                     self.Reachable(instr),
                                     "%s <==> reachable %s" %
                                     (str(self.statement_enablers),
                                      str(instr)
                                      )
                                     )
        return instr

    def _handle_globals_in_read(self,
                                instr: "Instruction",
                                expression: BaseExpression):
        assert len(expression.read_global_variables), \
            "After simplification, expression must contain exactly one global variable"
        for varname in expression.read_global_variables:
            val = self.make_global_read(varname, instr)
            self.local_var_table[varname] = val

    @methoddispatch
    def compile_statement(self, statement: Statement):
        raise NotImplementedError("no support for %s" % str(statement))

    @compile_statement.register(EmptyStatement)
    def _empty(self, statement: EmptyStatement):
        return

    @compile_statement.register(SimpleStatement)
    def _simple(self, statement: SimpleStatement):
        assert not statement.expression.read_global_variables, \
            "simple statement must not contain globals after simplification"

    @compile_statement.register(PrintlnStatement)
    def _println(self, println: PrintlnStatement):
        assert not any(expression.read_global_variables for expression in println.expressions), \
            "println statement must not contain global reads after simplification"

    @compile_statement.register(MultiStatement)
    def _multi(self, statement: MultiStatement):
        for statement in statement.statements:
            self.compile_statement(statement)

    @compile_statement.register(ITE)
    def _ite(self, ite: ITE):
        assert not ite.condition.read_global_variables, \
            "Condition must not contain global reads after simplification"
        expr = self.compile_expression(ite.condition)
        self.statement_enablers.append(expr)
        old_parents = self.parent_instructions
        consequence_parents = []
        alternative_parents = []
        self.compile_statement(ite.consequence)
        consequence_parents = self.parent_instructions
        self.parent_instructions = old_parents
        if ite.alternative:
            # flip last condition
            self.statement_enablers[-1] = Not(self.statement_enablers[-1])
            self.compile_statement(ite.alternative)
            alternative_parents = self.parent_instructions
        self.statement_enablers.pop()
        self.parent_instructions = consequence_parents + alternative_parents

    @compile_statement.register(VariableDeclaration)
    def _vardec(self, vardec: VariableDeclaration):
        # WARNING: This only handles local variable declarations
        var = self.create_local_pseuco_variable(vardec.name, vardec.var_type)
        if (vardec.initial_value.read_global_variables):
            instr = self._create_and_order_instruction(vardec)
            self._handle_globals_in_read(instr, vardec.initial_value)
        value = self.compile_expression(vardec.initial_value)
        fact = var == value
        # self.solver.add(fact)
        self.solver.assert_and_track(fact, str(fact))

    @compile_statement.register(VariableAssignment)
    def _varassign(self, assignment: VariableAssignment):
        assert assignment.is_global_write, "Unsupported modification of local"
        # TODO: cannot handle read and write of global in same statement
        assert not assignment.assignment_expression.read_global_variables, \
            "assignment must not contain global reads after simplification"
        written_value = self.compile_expression(
            assignment.assignment_expression)
        instr = self._create_and_order_instruction(assignment)
        self.make_global_write(assignment.var_name, instr, written_value)

    def _track_predecessors(self):
        for instruction in self._instruction2instruction_id:
            predecessor = z3.Const("pre"+str(instruction), self.Instruction)
            pre, notpre = z3.Consts("pre notpre", self.Instruction)
            clause = Implies(
                self.Reachable(instruction),
                Or(
                    ForAll([pre],
                           Implies(
                               self.InstructionOrder(pre, instruction),
                               Or(
                                   pre == instruction,
                                   Not(self.Reachable(pre)),
                                   self.context
                               )
                           )
                           ),
                    Exists([pre],
                           And(
                               self.InstructionOrder(pre, instruction),
                               self.Reachable(pre),
                               pre != instruction,
                               predecessor == pre,
                               Not(Exists([notpre],
                                          And(
                                              self.Reachable(notpre),
                                              notpre != pre,
                                              notpre != instruction,
                                              self.InstructionOrder(notpre, instruction),
                                              self.InstructionOrder(pre, notpre),
                                              self.context)
                                          )
                                   ),
                               self.context
                           ),
                        ),
                    self.context
                )
            )
            self.solver.add(clause)
            self.direct_predecessors[instruction] = predecessor

    def _extract_write_seen_from_model(self, model) -> List[Dict[str, InstructionId]]:
        """Extracts values from ReadWrite"""
        ret = []
        for read, *_ in self.global_reads:
            if z3.is_true(model.evaluate(self.Reachable(read))):
                write = model.evaluate(self.ReadWrite(read))
                ret.append({
                    "write": self._instruction2instruction_id[write],
                    "seen": self._instruction2instruction_id[read]
                })
        return ret

    def _extract_predecessors_from_model(self, model) -> List[Dict[str, InstructionId]]:
        """Converts direct_predecessors from an instruction mapping to a statement mapping"""
        ret = []
        for instr, instr_id in self._instruction2instruction_id.items():
            if instr_id.agent_id == 0 or z3.is_false(model.evaluate(self.Reachable(instr))):
                # initial statements do not have a predecessor
                continue
            pred_instr = model[self.direct_predecessors[instr]]
            pred_id = self._instruction2instruction_id[pred_instr]
            ret.append({
                "statement": instr_id,
                "predecessor": pred_id
            })
        return ret

    def _extract_reachable_from_model(self, model):
        return [instr_id for instr, instr_id in self._instruction2instruction_id.items()
                if z3.is_true(model.evaluate(self.Reachable(instr)))]

    def _assert_reads_and_writes(self):
        """Adds all reads and writes to the solver"""
        a = z3.Const('a', self.Instruction)
        x = z3.Const('x', self.PseuCoVar)
        p = z3.Const('p', z3.IntSort(ctx=self.context))
        # specify read and write function
        reads = ForAll([a, x, p],
                       self.ReadValue(a, x, p) == (
                           Or(*(And(a == b, x == y, p == q, self.context)
                                for (b, y, q)
                                in self.global_reads), self.context))
                       )
        writes = ForAll([a, x, p],
                        self.WriteValue(a, x, p) == (
                            Or(*(And(a == b, x == y, p == q, self.context)
                                 for (b, y, q)
                                 in self.global_writes), self.context))
                        )
        self.solver.assert_and_track(reads, "reads")
        self.solver.assert_and_track(writes, "writes")

    def _find_new_model(self, model, counter):
        current_model = (value == model[value].as_long()
                         for instr, _, value in self.global_reads
                         if z3.is_true(model.evaluate(self.Reachable(instr))))
        self.solver.assert_and_track(Not(And(*current_model, self.context)),
                                        "counter %s" % counter,)
        return self.solver.check()

    def _add_seqconsist_axioms(self):
        """Creates the axioms of sequential consistency"""
        logging.info("adding sequential consistency axioms")
        a, b, c = z3.Consts('a b c', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self.context))

        seqconsist_axioms = {
            "total": ForAll([a, b], Implies(
                And(self.Reachable(a), self.Reachable(b), self.context),
                Or(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self.context)
            )),  # order is total on reachable statements
            "transitive": ForAll([a, b, c], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, c), self.context),
                self.InstructionOrder(a, c)
            )),  # transitive
            "antisymmetric": ForAll([a, b], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self.context),
                a == b
            )),
            "write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self.context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                self.ReadWrite(a) == b,
                                self.InstructionOrder(b, a),
                                Not(
                                    Exists([c, q], And(
                                        self.Reachable(c),
                                        c != b,
                                        self.InstructionOrder(b, c),
                                        self.InstructionOrder(c, a),
                                        self.WriteValue(c, x, q),
                                        self.context
                                    ))
                                ),
                                self.context
                                ))
            )),
        }
        for name, fact in seqconsist_axioms.items():
            self.solver.assert_and_track(fact, name)

    def _add_weak_axioms(self):
        """Creates the axioms of sequential consistency"""
        logging.info("adding weak consistency axioms")
        a, b, c = z3.Consts('a b c', self.Instruction)
        x, y = z3.Consts('x y', self.PseuCoVar)
        p, q = z3.Consts('p q', z3.IntSort(ctx=self.context))


        weak_order_axioms = {
            "reflexive": ForAll([a] , Implies(self.Reachable(a),
                                self.InstructionOrder(a, a))),  # order is reflexive
            "transitive": ForAll([a, b, c], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, c), self.context),
                self.InstructionOrder(a, c)
            )),  # transitive
            "antisymmetric": ForAll([a, b], Implies(
                And(self.InstructionOrder(a, b), self.InstructionOrder(b, a), self.context),
                a == b
            )),
            "write-seen": ForAll([a, x, p], Implies(
                And(self.ReadValue(a, x, p), self.Reachable(a), self.context),
                Exists([b], And(self.WriteValue(b, x, p),
                                self.Reachable(b),
                                self.ReadWrite(a) == b,
                                Not(self.InstructionOrder(a, b)),
                                Not(
                                    Exists([c, q], And(
                                        self.Reachable(c),
                                        c != b,
                                        self.InstructionOrder(b, c),
                                        self.InstructionOrder(c, a),
                                        self.WriteValue(c, x, q),
                                        self.context
                                    ))
                                ),
                                self.context
                                ))
            )),
        }
        for name, fact in weak_order_axioms.items():
            self.solver.assert_and_track(fact, name)

    def compile(self, program):
        # 1. create the Instruction datatype/enum . It contains one constructor per instruction,
        # that is per statement which does a global read or write. Those consist of every variable
        # declaration in the initialization section, as well as writes and reads in each started
        # agent
        # Furthermore create a datatype for global variables
        # First, Instruction
        Instruction = z3.Datatype("Instruction", ctx=self.context)
        for declaration_counter in range(len(program.initialization.variable_declarations)):
            Instruction.declare("a0_%d" % declaration_counter)
        startedAgents = program.startedAgentNames.asList()
        for agent_counter, agent_name in enumerate(chain.from_iterable(startedAgents)):
            agent = program.initialization.agent_declarations[agent_name]
            self.agent_id2agent[agent_counter] = agent
            for statement_counter in range(agent.global_operations):
                Instruction.declare("a%d_%d" % (agent_counter+1, statement_counter))
        self.Instruction = Instruction.create()

        # Now, PseuCoVar
        PseuCoVar =  z3.Datatype('PseuCoVar', ctx=self.context)
        # every legal PseuCo variable name works as a python identifier, we can make use of that
        # they are also necessarily unique
        for declaration in program.initialization.variable_declarations:
            PseuCoVar.declare(declaration.name)
        self.PseuCoVar = PseuCoVar.create()

        # 2. Now that Instructions do exist, define relations on them
        self.InstructionOrder = z3.Function("InstructionOrder", self.Instruction, self.Instruction,
                                            z3.BoolSort(ctx=self.context))
        # Todo: support more than int for read and write
        self.ReadValue = z3.Function("read", self.Instruction, self.PseuCoVar,
                                     z3.IntSort(ctx=self.context), z3.BoolSort(ctx=self.context))
        self.WriteValue = z3.Function("write", self.Instruction, self.PseuCoVar,
                                      z3.IntSort(ctx=self.context), z3.BoolSort(ctx=self.context))
        # "output" relation, used to get the read-write-relation from the model
        self.ReadWrite = z3.Function("ReadWrite", self.Instruction, self.Instruction)

        # Many properties only have to hold for statements which are actually reachable
        # thus we need to define a predicate for reachability
        self.Reachable = z3.Function("reachable", self.Instruction, z3.BoolSort(ctx=self.context))

        # 3. generate constraints for global variables and their initialization
        for counter, declaration in enumerate(program.initialization.variable_declarations):
            instr = self.Instruction.__getattribute__("a0_%d" % counter)
            self._instruction2instruction_id[instr] = InstructionId(0, declaration.id)
            self.solver.assert_and_track(self.Reachable(instr), "reach %s" % str(instr))
            self.parent_instructions.append(instr)
            expr = self.compile_expression(declaration.initial_value)
            self.make_global_write(declaration.name, instr, expr)
        initial_parents = self.parent_instructions

        # 4. Iterate over all started agents, and generate code for them
        for agent_name in chain.from_iterable(startedAgents):
            self.parent_instructions = initial_parents
            self.current_agent_id += 1
            self._current_agent_used_id_counter = 0
            self.local_var_table = {}  # reset local variables
            agent = program.initialization.agent_declarations[agent_name]
            body = agent.functionBody
            for statement in body.statements:
                self.compile_statement(statement)
        self._assert_reads_and_writes()
        self._track_predecessors()
        sat_stat = self.solver.check()
        assert sat_stat == z3.sat, sat_stat

    def list_executions(self):
        logging.info("listing all executions")
        logging.info(self.solver)
        is_sat = self.solver.check()
        if is_sat == z3.unknown:
            if self.solver.reason_unknown() == 'timeout':
                logging.warn(self.solver.statistics())
                raise ResourceWarning("Timeout occured")
            return "Cannot decide given program. WTH did you do?"
        assert  is_sat != z3.unsat, "Program without constraints should always be sat or unknown"
        counter = 0
        reply = []
        self._push()
        while is_sat == z3.sat and counter < 50:
            logging.info(f"{counter}")
            counter += 1
            model = self.solver.model()
            execution = []
            for (instr, var, value) in self.global_reads:
                read_instr_id = self._instruction2instruction_id[instr]
                if z3.is_false(model.evaluate(self.Reachable(instr))):
                    continue  # skip unreachable reads
                read_value = model[value].as_long()
                execution.append({
                    "agent_statement_id": read_instr_id,
                    "read_value": read_value,
                    "read_var": str(var),
                })
            reply.append({
                "reads": execution,
                "ordering": self._extract_predecessors_from_model(model),
                "reachable": self._extract_reachable_from_model(model),
                "writeSeen": self._extract_write_seen_from_model(model),
            })
            is_sat = self._find_new_model(model, counter)
        self._pop()
        if counter == 50:
            print("too many reads!?!")
        return reply

    def _replay_manually_added_assertions(self) -> Tuple[bool, int]:
        logging.info("replaying assertions")
        self._manually_added_assertions, assertions = [], self._manually_added_assertions
        for count, assertion in enumerate(assertions):
            sat = self.add_constraint(assertion)
            if not sat:
                return False, count
        return True, 0

    def set_memory_model(self, memory_model: MemoryModel) -> Tuple[bool, int]:
        self._pop_all()
        self._push()
        if memory_model is MemoryModel.SequentialConsistency:
            self._add_seqconsist_axioms()
        elif memory_model is MemoryModel.WeakConsistency:
            self._add_weak_axioms()
        else:
            raise KeyError("unkown memory model: {memory_model}")
        return self._replay_manually_added_assertions()


    Assertion = Union[OrderAssertion, WriteSeenAssertion, ReadValueAssertion]
    def add_constraint(self, constraint: Assertion) -> bool:
        """addds an constraint if satisfiable and returns it sat status"""
        self._push()
        if isinstance(constraint, WriteSeenAssertion):
            write = self._instruction2instruction_id.inv[constraint.writer]
            read = self._instruction2instruction_id.inv[constraint.reader]
            self.solver.add(self.Reachable(write))
            self.solver.add(self.Reachable(read))
            anyval = z3.Int("anyval", self.context);
            var = self.PseuCoVar.__getattribute__(constraint.variable)
            rw = Exists([anyval], And(
                self.ReadValue(read, var, anyval),
                self.WriteValue(write, var, anyval),
                self.ReadWrite(read) == write,
                self.context
            ))
            self.solver.add(rw)
            self.solver.add(self.ReadWrite(read) == write)
        elif isinstance(constraint, ReadValueAssertion):
            instr = self._instruction2instruction_id.inv[constraint.instruction]
            var = self.PseuCoVar.__getattribute__(constraint.variable)
            value = z3.IntVal(constraint.value, self.context)
            self.solver.add(self.Reachable(instr))
            self.solver.add(self.ReadValue(instr, var, value))
        sat = self.solver.check() == z3.sat
        if not sat:
            self._pop()
        else:
            self._manually_added_assertions.append(constraint)
        return sat

    def remove_last_constraint(self):
        self._pop()
